/*
* COMPONENT_NAME: ldap.clients
*
* ABSTRACT: generic program to delete an entry using LDAP
*
* ORIGINS: 202,27
*/ 
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* ldapdelete.c - simple program to delete an entry using LDAP */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <ldap.h>
#include <ldapssl.h>
#include <ldif.h>

/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

#define DEFHOST "localhost"

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );

const char* ssl_err2string( int );

#ifndef _WIN32
#ifdef _GCC3
#include <errno.h>
#else
extern int errno;
#endif
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = (char *) "";
static char *ldaphost = (char *)DEFHOST;
static char *mech = NULL;
static int hoplimit = 10;
static int ldapport = LDAP_PORT;
static int doit = LDAP_TRUE;
static int verbose = LDAP_FALSE;
static int contoper = LDAP_FALSE;
static int referrals = LDAP_OPT_ON;
static LDAP *ld;
static int deref = LDAP_DEREF_NEVER;
static int ldapversion = LDAP_VERSION3;
static FILE *fp = NULL;
static int DebugLevel = 0;
static int keys = 0;
static int tls = 0;
static int ssl = 0;
static int use_FIPS_mode = 0; /* 0=off 1=on */
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_label = NULL;
static char *charset = NULL;
static char *debug_ptr = NULL;
extern char *optarg;
extern int optind;
LDAPControl **bindControls = NULL;
LDAPControl **opControls = NULL;
static int doNotReplicate = 0;

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends delete requests to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s [options] [DNs]\n", prog);
   fprintf(stderr, "    %s [options] [-i file]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "      dn: one or more items to delete\n");
   fprintf(stderr, "    file: name of input file containing items to delete\n");
   fprintf(stderr, "note:\n");
   fprintf(stderr, "    if neither dn or file is specified then items are read from standard input\n");
   fprintf(stderr, "options:\n" );
   fprintf(stderr, "    -c           continuous operation; do not stop processing on error\n");
   fprintf(stderr, "    -C charset   character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level     set debug level in LDAP library\n");
   fprintf(stderr, "    -D dn        bind dn\n");
   fprintf(stderr, "    -f file      read dn's from a file for deletion, one dn per line\n");
   fprintf(stderr, "    -G realm     realm used for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -h host      LDAP server host name\n");
   fprintf(stderr, "    -i file      read dn's from a file for deletion, one dn per line\n");
   fprintf(stderr, "    -k           use server administration control\n");
   fprintf(stderr, "    -K keyfile   file to use for keys\n");
   fprintf(stderr, "    -l           do not replicate the entry\n");
   fprintf(stderr, "    -m mechanism perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M           Manage referral objects as normal entries.\n");
   fprintf(stderr, "    -n           show what would be done but don't actually do it\n");
   fprintf(stderr, "    -N key_name  private key name to use in keyfile\n");
   fprintf(stderr, "    -O maxhops   maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port      LDAP server port number\n");
   fprintf(stderr, "    -P key_pw    keyfile password\n");
   fprintf(stderr, "    -R           do not chase referrals\n");
   fprintf(stderr, "    -s           delete subtree \n");
   fprintf(stderr, "    -U username  user name for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -v           verbose mode\n");
   fprintf(stderr, "    -V version   LDAP protocol version (2 or 3; default is 3)\n");
   fprintf(stderr, "    -w password  bind password or '?' for non-echoed prompt\n");
   fprintf(stderr, "                 use backslash '\\?' to avoid matching single character filenames (UNIX only)\n" );
   fprintf(stderr, "    -x           use FIPS mode processing (SSL/TLS only)\n");
   fprintf(stderr, "    -y proxydn   set proxied id for proxied authorization operation\n");
   fprintf(stderr, "    -Y           use a secure LDAP connection (TLS)\n");
   fprintf(stderr, "    -Z           use a secure LDAP connection (SSL)\n");
   exit( 1 );
}

/*
**      Print a memory error message and exit.
*/

void mem_error(const char *varName) {
   fprintf(stderr,"%s: Memory allocation error occured.\n", varName);
   exit(1);
}

/*
** Parse command line arguments.
*/

void parse_arguments(int argc, char **argv) {
   int i, port = 0, rc = LDAP_SUCCESS;
   int manageDsa = LDAP_FALSE;
   int adminControl = LDAP_FALSE;
   int subtreeControl = LDAP_FALSE;
   int Proxy_control_requested = LDAP_FALSE;
   char *proxyDN;
   int stringLen = 0;

   while ((i = getopt(argc, argv, "nvRMxYZcksl?h:V:p:D:w:d:i:f:m:O:K:P:N:C:y:G:U:")) != EOF) {
      switch ( i ) {
      case 'V':   /* version */
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf( stderr, "Invalid version level supplied.\n");
            usage();
         }
         break ;
      case 'O':         /* set maximum referral hop count  */
         hoplimit = atoi( optarg );
         break;
      case 'm':         /* use SASL bind mechanism  */
         if (!(mech = strdup ( optarg )))
            mem_error("mech");
         break;
      case 'c':   /* continuous operation mode */
         ++contoper;
         break;
      case 'h':   /* ldap host */
         if (!(ldaphost = strdup( optarg )))
            mem_error("host");
         break;
      case 'D':   /* bind DN */
         if (!(binddn = strdup( optarg )))
            mem_error("bindDN");
         break;
      case 'w':   /* password */
         if (optarg && optarg[0] == '?') {
            passwd = getPassword();

         } else
            if (!(passwd = strdup( optarg )))
               mem_error("password");
         break;
      case 'i':   /* read DNs from a file */
      case 'f':   /* read DNs from a file */
         if (( fp = fopen( optarg, "r" )) == NULL ) {
            perror( optarg );
            exit( 1 );
         }
         break;
      case 'd':   /* debug */
         DebugLevel = atoi( optarg );
         break;
      case 'p':   /* port */
         ldapport = atoi( optarg );
         port = 1;
         break;
      case 'n':   /* print deletes, don't actually do them */
         doit = LDAP_FALSE;
         break;
      case 'R':   /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'l':        /* don't replicate the entry after it's modified */
         doNotReplicate = 1;
         if ((rc = ldap_add_control (LDAP_REPL_DO_NOT_REPLICATE_CONTROL_OID, 0 ,
                                     NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modify: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'M':   /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         /*Add the control for the bind*/
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         /*Add the control for the delete operation*/
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'v':   /* verbose mode */
         verbose++;
         break;
      case 'K':   /* keyfile */
         if (!(keyfile = strdup( optarg )))
            mem_error("keyfile name");
         keys = 1;
         break;
      case 'P':   /* keyfile password */
         if (!(keyfile_pw = strdup( optarg )))
            mem_error("keyfile password");
         break;
      case 'N':   /* keyfile label */
         if (!(keyfile_label = strdup( optarg )))
            mem_error("keyfile label");
         break;
      case 'x':  /* enable FIPS mode*/
	 use_FIPS_mode = 1;
         break;
      case 'Y':   /* TLS */
         if (ssl) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         tls = 1;
         break;
      case 'Z':   /* SSL */
         if (tls) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         ssl = 1;
         break;
      case 'C':   /* charset */
         charset = strdup(optarg);
         break;
      case 'k':   /* admin control */
         adminControl = LDAP_TRUE;
         /*Add the control for the bind*/
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &bindControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         /*Add the control for the delete operation*/
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 's':   /* subtree control */
         subtreeControl = LDAP_TRUE;
         /*Add the control for the delete operation*/
         if ((rc = ldap_add_control (LDAP_DELETE_SUBTREE_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &opControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_delete: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'y':   /*   proxied authorization control */
          rc=ldap_proxy_dn_prefix(&proxyDN,optarg);
          if (rc == LDAP_SUCCESS) {
                stringLen=strlen(proxyDN);
                rc = ldap_add_control (LDAP_PROXY_CONTROL_OID,stringLen,
                      proxyDN,1,&opControls);
               free(proxyDN);
              }
          if (rc != LDAP_SUCCESS) {
              fprintf(stderr, "Incorrect Proxied DN specified. \n");
              usage();
               }
           else  Proxy_control_requested=LDAP_TRUE;
         break;
      case 'U':  /* DIGEST-MD5 Username */
         if ( !(digest_username = strdup( optarg )) ) {
            mem_error( "Username" );
         }
         break;
      case 'G':  /* DIGEST-MD5 Realm */
         if ( !(digest_realm = strdup( optarg )) ) {
            mem_error( "realm" );
         }
         break;
      case '?':
      default:
         usage();
      }
   }

   if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
      ssl = 1;                  /* Use SSL by default */

   if ( (manageDsa | subtreeControl | adminControl | Proxy_control_requested) && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M,-s and -k options require LDAP version 3. ");
      if (manageDsa)
         fprintf(stderr, "-M ignored. ");
      if (subtreeControl)
         fprintf(stderr, "-s ignored. ");
      if (adminControl)
         fprintf(stderr, "-k ignored.");
      fprintf(stderr, "\n\n");
   }


   /* set to standard file input if no additional arguments specified */
   if ((fp == NULL) && (optind >= argc))
      fp = stdin;

   /* Use default SSL port if none specified */
   if ((port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT;

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );

   /*Set up the controls*/
   /*Add the password policy controls for the bind and the delete operation*/
   if (ldapversion == LDAP_VERSION3) {
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &bindControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
   }
   if (ldapversion == LDAP_VERSION3) {
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &opControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }
   }
}

/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   fputs( "Enter password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}

/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
         *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/

void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **bind_client_controls = NULL;

   /* Set FIPS mode processing if requested for TLS and SSL. */
   if (( tls == 1 || ssl == 1 ) && use_FIPS_mode == 1 ) {
      if (verbose)
	 printf( "ldap_ssl_set_fips_mode_np(1)\n");
      
      rc = ldap_ssl_set_fips_mode_np( 1 ); 
      if (rc != LDAP_SUCCESS) {
	 fprintf( stderr,
		  "ldap_ssl_set_fips_mode_np failed\n" );
	 exit( rc ) ;
      }
   }

   /* call ldap_ssl_client_init if V3 and SSL */
   if ((tls || ssl) && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         fprintf( stderr, "%s\n", ssl_err2string( failureReasonCode ));
         exit( 1 ) ;
      }
   }

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport,
                   ((keyfile_label) ? keyfile_label : "NULL"));
         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_label ) ;
         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport);
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
            perror( "ldap_init" );
            exit(1);
         }
      }
   } else {
      if ( verbose )
         printf("ldap_open( %s, %d )\n",
                ((ldaphost) ? ldaphost : "NULL"), ldapport );
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )(long)DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )(long)deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)(long)hoplimit);
   }
   /* don't chase referrals during the bind */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_OFF);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "Unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }

   /* call ldap_ssl_start if V2 and SSL */
   if ( (tls || ssl) && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_label) ? keyfile_label : "NULL"));

      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_label);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   if (tls) {                   /* Start the TLS session */
      if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
         fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_start_tls_s_np: %s\n", errormsg);
         exit(rc);
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {
      if ( ! mech ) /* Use simple bind */
      {
         ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, bindControls, NULL, &msgidp);
         rc2 = rc; /* Store this result code for later user. */
         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                                    NULL, &returnedControls, 1 );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, bindControls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               ldap_get_option( ld, LDAP_OPT_EXT_GSS_ERR, &failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Major Status: %i\n", failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Minor Status: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else /* other SASL mechanisms */ {
            if (strcmp(mech,LDAP_MECHANISM_DIGEST_MD5) == 0) {
               if (digest_username) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) ,        digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
               if (digest_realm) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) ,  digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
            }

            ber.bv_len = strlen ( passwd );
            ber.bv_val = passwd;
            rc = ldap_sasl_bind_s (ld, binddn, mech, &ber,
                                   bindControls, bind_client_controls,
                                   &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         }

         /* Get any controls sent by the server during the sasl_bind_s */
         rc2 = rc;

         if ( rc == LDAP_SUCCESS ) {
            rc = ldap_get_bind_controls( ld, &returnedControls );
         }
      }

      /*Free the client controls*/
      ldap_controls_free(bind_client_controls);

      if (rc == LDAP_SUCCESS) {
         if (returnedControls != NULL) {
            ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
         }
      }
      if (rc2 != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc2);
      } else if (rc != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc);
      } else if (errcode != LDAP_SUCCESS) {
         errormsg = ldap_err2string(errcode);
      }
      if (controlerr) {
         errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
      }
      if (controlwarn) {
         warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
      }
      if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
         fprintf(stderr,"ldap_simple_bind:  ");
         if (errormsg != NULL) {
            fprintf(stderr,errormsg);
            fprintf(stderr," --- ");
         }
         if (errormsgctrl != NULL) {
            fprintf(stderr,errormsgctrl);
            fprintf(stderr,"\n");
         }
         if (warnmsgctrl != NULL) {
            fprintf(stderr,warnmsgctrl, controlres);
            fprintf(stderr,"\n");
         }
      } else if (errormsg != NULL) {
         fprintf(stderr,"ldap_simple_bind:  %s \n", errormsg);
      }
      if (rc2 != LDAP_SUCCESS) {
         /*
         ** ldap_get_exterror only valuable for ssl communication.
         ** In this example, for LDAP v3, the bind is the first
         ** instance in which communication actually flows to the
         ** server.  So, if there is an ssl configuration error or
         ** other ssl problem, this will be the first instance where
         ** it will be detected.
         */
         if (ssl) {
            fprintf( stderr, "Attempted communication over SSL.\n");
            fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
         }
         exit( rc2 );
      } else if (errcode != LDAP_SUCCESS) {
         exit( errcode );
      }

   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }

   /* during the operation, use user setting to chase referrals */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )(long)referrals);
}

/* ssl_err2string returns a string describing the error associated with
** the given error code from an SSL related call.
*/
const char * ssl_err2string( int ssl_error ) {
   switch ( ssl_error ) {
   case LDAP_SSL_EXTENDED_OK:
      return LDAP_SSL_EXTENDED_OK_STR;
   case LDAP_SSL_EXTENDED_INVALID_HANDLE:
      return LDAP_SSL_EXTENDED_INVALID_HANDLE_STR;
   case LDAP_SSL_EXTENDED_API_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_API_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_INTERNAL_ERROR:
      return LDAP_SSL_EXTENDED_INTERNAL_ERROR_STR;
   case LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE:
      return LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE_STR;
   case LDAP_SSL_EXTENDED_INVALID_STATE:
      return LDAP_SSL_EXTENDED_INVALID_STATE_STR;
   case LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND:
      return LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND_STR;
   case LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION:
      return LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_CRYPTO:
      return LDAP_SSL_EXTENDED_ERROR_CRYPTO_STR;
   case LDAP_SSL_EXTENDED_ERROR_ASN:
      return LDAP_SSL_EXTENDED_ERROR_ASN_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR:
      return LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR_STR;
   case LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR:
      return LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR:
      return LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT:
      return LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL_STR;
   case LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD:
      return LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED:
      return LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB:
      return LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB_STR;
   case LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR:
      return LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR_STR;
   case LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR:
      return LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR_STR;
   case LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR:
      return LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_INDEX:
      return LDAP_SSL_EXTENDED_ERROR_BAD_INDEX_STR;
   case LDAP_SSL_EXTENDED_CLOSE_FAILED:
      return LDAP_SSL_EXTENDED_CLOSE_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_DATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_DATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS:
      return LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_IO:
      return LDAP_SSL_EXTENDED_ERROR_IO_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MAC:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MAC_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_PEER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_PEER_STR;
   case LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED:
      return LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED_STR;
   case LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED:
      return LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED:
      return LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY:
      return LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER:
      return LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER_STR;
   case LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE:
      return LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE_STR;
   case LDAP_SSL_EXTENDED_WOULD_BLOCK:
      return LDAP_SSL_EXTENDED_WOULD_BLOCK_STR;
   case LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3:
      return LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3_STR;
   case LDAP_SSL_EXTENDED_MISC_INVALID_ID:
      return LDAP_SSL_EXTENDED_MISC_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE_STR;
   case LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING:
      return LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING_STR;
   default:
      return "Unknown SSL error";
   }
}


/*
** Delete the item or subtree as per requirement
*/

int dodelete(LDAP *ld, char *dn) {
   int rc;
   int rc2 = 0, error = 0, errcode = 0;
   LDAPMessage *res;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   int msgidp = 0;
   LDAPControl **returnedControls = NULL;

   printf( "%sDeleting entry %s\n", doit ? "" : "!", dn );
   if (!doit)
      return LDAP_SUCCESS;

   rc = ldap_delete_ext( ld, dn, opControls, NULL, &msgidp );

   rc2 = rc; /* Store this result code for later user. */
   if (rc == LDAP_SUCCESS) {
      error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
      rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                              NULL, &returnedControls, 1 );
   }
   if (rc == LDAP_SUCCESS) {
      if (returnedControls != NULL) {
         ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
      }
   }
   if (rc2 != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc2);
      rc = rc2;
   } else if (rc != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc);
   } else if (errcode != LDAP_SUCCESS) {
      errormsg = ldap_err2string(errcode);
      rc = errcode;
   }

   if (rc == LDAP_SERVER_DOWN) {
      exit(rc);
   }

   if (controlerr) {
      errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
   }
   if (controlwarn) {
      warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
   }
   if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
      fprintf(stderr,"ldap_delete:  ");
      if (errormsg != NULL) {
         fprintf(stderr,errormsg);
         fprintf(stderr," --- \n");
      }
      if (errormsgctrl != NULL) {
         fprintf(stderr,errormsgctrl);
         fprintf(stderr,"\n");
      }
      if (warnmsgctrl != NULL) {
         fprintf(stderr,warnmsgctrl, controlres);
         fprintf(stderr,"\n");
      }
   } else if ((errormsg != NULL) || (errmsg != NULL )) {
      if (errormsg != NULL)
         if (*errormsg != '\0')
            fprintf(stderr,"ldap_delete:  %s \n", errormsg);
      if (errmsg != NULL)
         if (*errmsg != '\0')
            fprintf(stderr,"ldap_delete: additional info: %s \n", errmsg);
   } else if ( verbose ) {
      printf( "entry removed\n" );
   }
   return ( rc );
}

int main( int argc, char **argv ) {
   int ctr = 0, rc = 0;
   char buf[4096];

   /* Strip off any path info on program name */
#if defined( WIN32 ) || defined( _WIN32 )

   if ((prog = strrchr(argv[0], '\\')) != NULL)
      ++prog;
   else
      prog = argv[0];
#else

   if (prog = strrchr(argv[0], '/'))
      ++prog;
   else
      prog = argv[0];
#endif

   /* parse command line arguments */
   parse_arguments(argc, argv);

   /* connect to server */
   if (doit)
      connect_to_server();

   ctr = 0;

   /* delete items */
   if ( fp == NULL ) {
      for ( ; optind < argc; ++optind ) {
         rc = dodelete(ld, argv[optind]);
      }
   } else {
      rc = 0;
      while ((rc == 0 || contoper) && fgets(buf, sizeof(buf), fp) != NULL) {
         buf[strlen( buf ) - 1] = '\0'; /* remove trailing newline */
         if (*buf != '\0')
            rc = dodelete(ld, buf);
      }
   }

   /* unbind and exit */
   if (doit) {
      if ( tls )
         ldap_stop_tls_s_np ( ld );
      else
         ldap_unbind(ld);
   }
   if (bindControls)
      ldap_controls_free(bindControls);
   if (opControls)
      ldap_controls_free(opControls);
   exit(rc);
}
