/*
* COMPONENT_NAME: ldap.audit
*
* ABSTRACT: Audit plugin header file
*
* ORIGINS: 202,27
*
* (C) COPYRIGHT International Business Machines Corp. 2000, 2004
* All Rights Reserved
* Licensed Materials - Property of IBM
*
* US Government Users Restricted Rights - Use, duplication or
* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*/

/*
 *
 * Licensed Materials - Property of IBM
 * 5647-A01,5722-SS1
 * (C) Copyright IBM Corp. 2000, 2004
 *
 */

#ifndef _AUDIT_PLUGIN_H
#define _AUDIT_PLUGIN_H

#include <ldap.h>

/** 
 * @file audit-plugin.h
 * @ingroup ITDS_Audit
*/

/** au_boolean defines a value of either    
 * AU_FALSE or AU_TRUE                     
 * @ingroup ITDS_Audit
*/
typedef unsigned char au_boolean;

/** au_ldapdn defines a character in        
 * local code page                         
 * @ingroup ITDS_Audit
*/
typedef char au_ldapdn;

#define AU_FALSE    0
#define AU_TRUE     1

#define AU_VER_UNKNOWN   0
#define AU_VER_2   1
#define AU_VER_3   2

#define AU_AUTH_UNKNOWN   0
#define AU_AUTH_ANONYMOUS  1
#define AU_AUTH_UNAUTHENTICATED  2
#define AU_AUTH_AUTHENTICATED  3

#define AU_NULL_DN   "<*CN=NULLDN*>"


#ifdef __OS400__
#define AU_NO_LOCAL_USER  "*NONE     "
#else
#define AU_NO_LOCAL_USER  "<*CN=NULLDN*>"
#endif

#define AU_YES                          'Y'
#define AU_NO                           'N'
#define AU_UNKNOWN                      'U'

/* The following lists the operation code for an     
 *  audit record.                                     
 *  EXTOP are for LDAP v3 extended operations.   */
#define AU_OP_UNKNOWN   0
#define AU_OP_BIND   1
#define AU_OP_UNBIND   2
#define AU_OP_SEARCH   3
#define AU_OP_ADD   4
#define AU_OP_MODIFY   5
#define AU_OP_DELETE   6
#define AU_OP_MODIFYDN   7
#define AU_OP_EXTOP   8
#define AU_OP_COMPARE                   9


#define AU_EXTOP_UNKNOWN  0
#define AU_EXTOP_EN_REGISTRATION 1
#define AU_EXTOP_EN_UNREGISTER  2
#define AU_EXTOP_GENERAL   3

/* --------------------------------------------
 * represents the data passed to audit plug-ins 
 * --------------------------------------------
 */
#define AU_DATA_CONFIG_INFO  0
#define AU_DATA_EVENT                   1
#define AU_DATA_RECORD   2

typedef struct _Audit_data {
   /** Audit config info, audit event or audit record*/
   int au_data_type;
   /** pointer to audit data       */
   void * au_data;
}
Audit_data;

/**Configuration for the Audit Control.*/
typedef struct _Audit_control_info {
	/**Max number of characters allowed a single UID.*/
	int max_num_characters;
	/**Max number of UIDs in a control.*/
	int max_num_UID;
	/**Allow additional info to be audited from non-admin.*/
	au_boolean allow_non_admin;
} Audit_control_info;


/**Number of Options in the Audit_config_info struct*/
#define NUM_OPTIONS              14

/** Structure to store the audit configuration options.
 *  If you add, or remove an option you must change the
 *  NUM_OPTIONS accordingly
*/
typedef struct _Audit_config_info {
   /** path/name of the log file */
   char * au_config_auditLog;
   /** Auditing enabled           */
   au_boolean au_config_audit;
   /**audit failed operations only  */
   au_boolean au_config_auditFailedOPonly;
   /**audit bind operations*/
   au_boolean au_config_auditBind;
   /**audit unbind operations*/
   au_boolean au_config_auditUnbind;
   /**audit search operations*/
   au_boolean au_config_auditSearch;
   /**audit add operations*/
   au_boolean au_config_auditAdd;
   /**audit modify operations*/
   au_boolean au_config_auditModify;
   /**audit delete operations*/
   au_boolean au_config_auditDelete;
   /**audit modify dn operations*/
   au_boolean au_config_auditModifyDN;
   /**Audit event notification extended operations*/
   au_boolean au_config_auditExtOPEvent;
   /**Audit all extended operations: not including event notification*/
   au_boolean au_config_auditExtOP;
   /**Audit all attributes sent on a group evaluation extended operation*/
   au_boolean au_config_auditAttrs;
   /**Audit compare operation*/
   au_boolean au_config_auditCompare;
   /**Audit groups sent on group control*/
   au_boolean au_config_auditGroups;
   /**Audit version*/
   int au_config_version;
   /**Audit Control configuration*/
   Audit_control_info au_config_control;
}
Audit_config_info;

/** Array version of the structure to store the
 *  audit configuration options.
 *  If you add, or remove an option you must change the
 *  NUM_OPTIONS accordingly
*/
typedef struct _Audit_config_info_a {
   /**Path/name of the audit log file*/
   char * au_config_auditLog;
   /**Array of audit configuration options*/
   au_boolean au_config_options[NUM_OPTIONS];
    /**Audit version*/
   int au_config_version;
   /**Audit Control configuration*/
   Audit_control_info au_config_control;
}
Audit_config_info_a;


/**Audit Event:  Auditing Started*/
#define AU_EVENT_AUDITING_STARTED 0 
/**Audit Event:  Auditing Ended*/
#define AU_EVENT_AUDITING_ENDED   1 
/**Audit Event:  Auditing Configuration Changed*/
#define AU_EVENT_CONFIG_OPTION_CHANGED 2 
/**Audit Event:  Auditing Server Started*/
#define AU_EVENT_SERVER_STARTED         3 
/**Audit Event:  Auditing Server Stopped*/
#define AU_EVENT_SERVER_ENDED           4

/**Structure to store an Audit Event*/
typedef struct _Audit_event {
   /**Time when the event occurred*/
   time_t au_event_time;
   /**Time's milliseconds*/
   int au_event_time_ms;
   /**Audit Event Number*/
   int au_event_number;
   /**Message associated with event
     * text will be in local code page*/
   char * au_event_message;
}
Audit_event;


/**Structure to store an audit record.*/
typedef struct _Audit_record {
   /**Time when operation processing is done*/
   time_t au_end_time;
   /**au_end_time's milliseconds*/
   int au_end_time_ms;
   /**Time when operation processing started*/
   time_t au_start_time;
   /**au_start_time's milliseonds*/
   int au_start_time_ms;
   /**LDAP version number */
   int au_version;
   /**SSL connection*/
   au_boolean au_ssl;
   /**Authentication type*/
   int au_auth;
   /**Connection ID*/
   int au_connid;
   /*Client IP Address*/
   char * au_clientIp;
   /**Client Port Number*/
   int au_clientPort;
   /**LDAP operation*/
   int au_operation;
   /**Extended operation if au_operation is AU_OP_EXTOP*/
   int au_extOP;
   /**Bind DN*/
   au_ldapdn * au_dn;
   /**Transaction ID*/
   char * au_transID;
   /**Result or status.  LDAP error code*/
   int au_rc;
   /**Pointer to control data, null if none.
    * This is a pointer to an array of pointers
    * which point to the LDAP control structures
    */
   LDAPControl ** au_control;
   /**Pointer to operation specific data.
    *Null if none.  Use au_opeation to determine which operation
    *structure to map over the data
    */
   void * au_specific;
   /**local user identity if known*/
   char * au_local_user;
   /**Administrator?
    * Y= administrator, N=not the adminsitrator, U=unknown*/
   char au_admin;
   /*TLS connection*/
   au_boolean au_tls;
   /**Unique ID for the operation*/
   char * au_uniqueID;
   /**Is the connection an admin connection?  This value 
    does not change even with the proxy auth control.*/
   int au_c_admin;
}
Audit_record;

/* == operation specific data ============================================*/

/* ---------------------------------
 * structure for Bind operation data
 * ---------------------------------
 */ 
/**Unknown authentication */
#define AU_AUTH_CHOICE_UNKNOWN  0 
/**Simple authentication*/
#define AU_AUTH_CHOICE_SIMPLE  1 
/**Kerberos v42 ldap*/
#define AU_AUTH_CHOICE_KRBV42LDAP 2 
/**Kerberos V42 DSA*/
#define AU_AUTH_CHOICE_KRBV42DSA  3 
/**SASL*/
#define AU_AUTH_CHOICE_SASL  4

/**Non-admin account */
#define AU_ACCT_NOT_ADMIN            0
/** Admin account not locked */
#define AU_ACCT_ADMIN_NOT_LOCKED     1   
/** Admin account locked */
#define AU_ACCT_ADMIN_LOCKED         2
/** Admin account lock cleared */
#define AU_ACCT_ADMIN_LOCK_CLEARED   3

/**Operation specific data structure for bind.
 */
typedef struct _Audit_data_bind {
   /**bind dn*/
   au_ldapdn * au_bind_name;
   /**authentication choice*/
   int au_bind_auth_choice;
   /**authentication mechanism*/
   char * au_bind_auth_mechanism;
   /**Username supplied on digest-md5 bind */
   char *au_bind_digest_username;
   /**Authzid supplied on digest-md5 bind*/
   char *au_bind_digest_authzid;
   /**Dn that username or authzid mapped to*/
   au_ldapdn *au_bind_digest_mapped_dn;
   /** Admin account lock status */
   int au_bind_admin_acct_lock_status;
}
Audit_data_bind;

/** There is no structure for the unbind operation.
 */

/**Search scope options*/
/**Unknown search scope*/
#define AU_SCOPE_UNKNOWN      0 
/**Search Scope: Base*/
#define AU_SCOPE_BASE         1 
/**Search Scope: one level*/
#define AU_SCOPE_ONELEVEL     2 
/**Search Scope: subtree*/
#define AU_SCOPE_SUBTREE      3

/**Search alias deref options*/
/**Deref Alias: Unknown*/
#define AU_DEREF_UNKNOWN      0 
/**Deref Alias: Never*/
#define AU_DEREF_NEVER        1 
/**Deref Alias: Searching*/
#define AU_DEREF_SEARCHING    2 
/**Deref Alias: Finding*/
#define AU_DEREF_FINDING      3 
/**Deref Alias: Always*/
#define AU_DEREF_ALWAYS       4

/**Operation specific data structure for search.
 */
typedef struct _Audit_data_search {
   /** base object entry      */
   au_ldapdn * au_search_baseObject;
   /**scope of search*/
   int au_search_scope;
   /** how alias is handled?  */
   int au_search_derefAliases;
   /** just return attr types?*/
   au_boolean au_search_typesOnly;
   /** search filter string   */
   char * au_search_filter;
   /** attributes to return,
   * attributes are listed  
   * one by one, and        
   * separated by comma (,) 
   */
   char * au_search_attributes;
}
Audit_data_search;

/**Operation specific data structure for add.
 */
typedef struct _Audit_data_add {
   /** entry to be added      */
   au_ldapdn * au_add_entry;
   /** attributes to add,
   * attributes are listed  
   * one by one, and        
   * separated by comma (,) 
   */
   char * au_add_attributes;
}
Audit_data_add;

/**Operation specific data structure for modify.
 */
typedef struct _Audit_data_modify {
   /** object to modify       */
   au_ldapdn * au_modify_object;
   /** changes to make    */
   LDAPMod ** au_modify_modification;

}
Audit_data_modify;

/**Modification Types*/
/**Modification Type: Unknown*/
#define AU_MOD_UNKNOWN        0 
/**Modification Type: Add*/
#define AU_MOD_ADD            1 
/**Modification Type: Delete*/
#define AU_MOD_DELETE         2 
/**Modification Type: Replace*/
#define AU_MOD_REPLACE        3


/**Operation specific data structure for delete.
 */
typedef struct _Audit_data_delete {
   /** entry to be deleted    */
   au_ldapdn * au_delete_entry;
}
Audit_data_delete;

/**Operation specific data structure for modify DN.
 */
typedef struct _Audit_data_modifyDN {
   /** entry to be changed    */
   au_ldapdn * au_modifyDN_entry;
   /** new rdn                */
   au_ldapdn * au_modifyDN_newrdn;
   /** delete old rdn?       */
   au_boolean au_modifyDN_deleteoldrdn;
   /** new immediate superior
    * NULL if none specified */
   au_ldapdn * au_modifyDN_newSuperior;
}
Audit_data_modifyDN;


/** List of the event IDs for Event         */
/* Notification                                        */
/**Event ID: unknown*/
#define AU_EXTOP_EN_ID_UNKNOWN   0 
/**Event ID: ldap change*/
#define AU_EXTOP_EN_ID_LDAP_CHANGE  1


/** ChangeTypes associated
 * with the LDAP Event Notification extended operation:
 * Registration request                                
 */ 
/**Change type: unknown*/
#define AU_EXTOP_EN_CHANGE_UNKNOWN 0 
/**Change type: add*/
#define AU_EXTOP_EN_CHANGE_ADD  1 
/**Change type: delete*/
#define AU_EXTOP_EN_CHANGE_DELETE 2 
/**Change type: modify*/
#define AU_EXTOP_EN_CHANGE_MODIFY 3 
/**Change type: moddn*/
#define AU_EXTOP_EN_CHANGE_MODDN 4

/**Operation specific data structure for LDAP Event Notification
 * extended operation: Registration Request.
 */
typedef struct _Audit_data_extOP_EN_registration {
   /**event id*/
   int au_extOP_EN_reg_eventID;
   /**base object*/
   au_ldapdn * au_extOP_EN_reg_baseObject;
   /**scope */
   int au_extOP_EN_reg_scope;
   /**type:
    * type are listed  
    * one by one, and  
    * separated by     
    * comma (,)        
    */
   char * au_extOP_EN_reg_type;
}
Audit_data_extOP_EN_registration;


/**Operation specific data structure for LDAP Event Notification
 * extended operation: Unregister Event
 */
typedef struct _Audit_data_extOP_EN_unregister {
   /**event registration id to unregister*/
   char * au_extOP_EN_unreg_ID;
}
Audit_data_extOP_EN_unregister;

/**Operation specific data structure for any Extended Operation
 */
typedef struct _Audit_data_extOP {
   /**Extended Operation oid*/
   char * au_extOP_oid;
   /**Additional info about the extended operation*/
   char * au_extOP_info;  /* additional info about extended op*/
   /**More additional information about extended ops.
    * attributes sent on a group eval exop
    */
   char * au_extOP_attrs;
}
Audit_data_extOP;

/**Operation specific data structure for compare.
 */
typedef struct _Audit_data_compare {
   au_ldapdn * au_compare_object;
   char * au_compare_attribute;
}
Audit_data_compare;

#endif /* _AUDIT_PLUGIN_H */
